#define _XOPEN_SOURCE 600
#include <stdlib.h>
#include <stdio.h>
#include <isds.h>
#include <string.h>
#include <readline/readline.h>
#include <readline/history.h>
#include <locale.h>
#include <unistd.h>
#include <string.h>
#include <stdint.h>
#include <confuse.h>
#include <errno.h>
#include <libgen.h>
#include <limits.h>

#include "shigofumi.h"
#include "completition.h"
#include "data.h"
#include "io.h"
#include "utils.h"

#define CONFIG_FILE ".shigofumirc"
#define CONFIG_SERVER "base_url"
#define CONFIG_USERNAME "username"
#define CONFIG_PASSWORD "password"
#define CONFIG_DEBUGLEVEL "debug_level"
#define CONFIG_VERIFYSERVER "verify_server"
#define CONFIG_CAFILE "ca_file"
#define CONFIG_CADIRECTORY "ca_directory"
#define CONFIG_CRLFILE "crl_file"
#define CONFIG_TIMEOUT "timeout"
#define CONFIG_LOGFACILITIES "log_facilities"
#define CONFIG_LOGFILE "log_file"
#define CONFIG_LOGLEVEL "log_level"
#define CONFIG_NORMALIZEMIMETYPE "normalize_mime_type"
#define CZPDEPOSIT_URL "https://www.czechpoint.cz/uschovna/"

#define TIMEOUT 10000
#define LOG_LEVEL 20

/* Configuration */
cfg_opt_t configuration_syntax[] = {
    CFG_STR(CONFIG_SERVER, (char *)isds_locator, CFGF_NONE),
    CFG_STR(CONFIG_USERNAME, NULL, CFGF_NODEFAULT),
    CFG_STR(CONFIG_PASSWORD, NULL, CFGF_NODEFAULT),
    /*CFG_STR(CONFIG_DEBUGLEVEL, NULL, CFGF_NODEFAULT),*/
    CFG_BOOL(CONFIG_VERIFYSERVER, cfg_true, CFGF_NONE),
    CFG_STR(CONFIG_CAFILE, NULL, CFGF_NODEFAULT),
    CFG_STR(CONFIG_CADIRECTORY, NULL, CFGF_NODEFAULT),
    CFG_STR(CONFIG_CRLFILE, NULL, CFGF_NODEFAULT),
    CFG_INT(CONFIG_TIMEOUT, TIMEOUT, CFGF_NONE),
    CFG_STR_LIST(CONFIG_LOGFACILITIES, "{none}", CFGF_NONE),
    CFG_STR(CONFIG_LOGFILE, NULL, CFGF_NODEFAULT),
    CFG_INT(CONFIG_LOGLEVEL, LOG_LEVEL, CFGF_NONE),
    CFG_BOOL(CONFIG_NORMALIZEMIMETYPE, cfg_true, CFGF_NONE),
    CFG_END()
};
cfg_t *configuration;

/* Logger */
int logger_fd = -1;

/* UI */
_Bool batch_mode = 0;
char *prompt = NULL;
struct command (*commands)[] = NULL;

/* Data */
struct isds_ctx *cisds = NULL;
char *server = NULL;
char *username = NULL;
char *password = NULL;
struct isds_list *boxes = NULL;
struct isds_message *message = NULL;
_Bool messages_are_outgoing = 0;
struct isds_list *messages = NULL;
unsigned long int total_messages = 0;
struct isds_ctx *czechpoint = NULL;

static void discard_credentials(void) {
    zfree(server);
    zfree(username);
    zfree(password);
}


/* Do the cleanup and exit */
static void shi_exit(int exit_code) {
    /* Data */
    discard_credentials();
    isds_list_free(&boxes);
    isds_message_free(&message);
    isds_list_free(&messages);

    if (cisds) {
        printf(_("Logging out...\n"));
        isds_logout(cisds);
        isds_ctx_free(&cisds);
    }
    isds_ctx_free(&czechpoint);
    isds_cleanup();
    
    /* Configuration */
    cfg_free(configuration);

    /* UI */
    free(prompt);
    free(commands);

    exit(exit_code);
}

/* Set prompt. if @format is NULL, switch to default prompt */
static void set_prompt(const char *format, ...) {
    char *buffer = NULL;
    va_list ap;
    if (format) {
        va_start(ap, format);
        shi_vasprintf(&buffer, format, ap);
        va_end(ap);

        if (buffer) {
            shi_asprintf(&prompt, _("%s> "), buffer);
            if (prompt) {
                free(buffer);
                return;
            }
        }

        free(buffer);
        free(prompt);
        prompt = strdup(_("> "));
        return;
    }

    zfree(prompt);
}


static int shi_load_configuration(const char *config_file) {
    char *config_name = NULL;
    int ret;

    /* Get config file */
    if (config_file) {
        config_name = (char *) config_file;
    } else {
        if (-1 == shi_asprintf(&config_name, "%s/%s", getenv("HOME"),
                    CONFIG_FILE)) {
            fprintf(stderr, _("Could not build configuration file name\n"));
            return -1;
        }
    }
    
    /* Parse configuration */
    configuration = cfg_init(configuration_syntax, CFGF_NONE);
    ret = cfg_parse(configuration, config_name);
    if (ret) {
        if (ret == CFG_FILE_ERROR) {
            fprintf(stderr,
                    _("Error while opening configuration file `%s': %s\n"),
                    config_name, strerror(errno));
        } else {
            fprintf(stderr, _("Error while parsing configuration file `%s'\n"),
                    config_name);
        }
        printf(_("Using default configuration\n"));
    }

    if (config_name != config_file) free(config_name);
    return 0;
}


static void finish_isds_operation(struct isds_ctx *ctx, isds_error err)  {
    shi_progressbar_finish();
    if (err) {
        if (isds_long_message(ctx)) 
            printf(_("Error occurred: %s: %s\n"), isds_strerror(err),
                    isds_long_message(ctx));
        else
            printf(_("Error occurred: %s\n"), isds_strerror(err));
    }
}


void logger(isds_log_facility facility, isds_log_level level,
        const char *message, int length, void *data) {
    int fd;
    ssize_t written, left = length;

    if (!data) return;
    fd = *((int *) data);
    /*printf("\033[32mLOG(%02d,%02d): ", facility, level);
    printf("%.*s", length, message);
    printf("\033[m");*/

    while (left) {
        written = write(fd, message + length - left, left);
        if (written == -1) {
            fprintf(stderr,
                    _("Could not save log message into log file: %s\n"
                        "Log message discarded!\n"),
                    strerror(errno));
            /*close(fd);
            fd = -1;*/
            return;
        }
        left-=written;
    }
}


/* Redirect ISDS log to file if @file is not NULL. */
static int do_log_to_file(const char *file) {
    if (file && *file) {
        logger_fd = open_file_for_writing(file, 0);
        if (logger_fd == -1) return -1;
        isds_set_log_callback(logger, &logger_fd);
    }
    return 0;
}


/* Add log facility based on its name. */
static int add_log_facility(isds_log_facility *facilities, const char *name) {
    if (!facilities) return -1;

    if (!strcmp(name, "none")) *facilities |= ILF_NONE;
    else if (!strcmp(name, "http")) *facilities |= ILF_HTTP;
    else if (!strcmp(name, "soap")) *facilities |= ILF_SOAP;
    else if (!strcmp(name, "isds")) *facilities |= ILF_ISDS;
    else if (!strcmp(name, "file")) *facilities |= ILF_FILE;
    else if (!strcmp(name, "sec")) *facilities |= ILF_SEC;
    else if (!strcmp(name, "xml")) *facilities |= ILF_XML;
    else if (!strcmp(name, "all")) *facilities |= ILF_ALL;
    else {
        printf(_("%s: Unknown log facility\n"), name);
        return -1;
    }

    return 0;
}


/* Clamp long int to unsigned int */
static unsigned int normalize_timeout(long int raw) {
    if (raw < 0) {
        printf(_("Configured network timeout is less then 0. Clamped to 0.\n"));
        return 0;
    }
    if (raw > UINT_MAX ) {
        printf(_("Configured network timeout is greater then %1$u. "
                    "Clamped to %1$u.\n"), UINT_MAX);
        return UINT_MAX;
    }
    return (unsigned int) raw;
}


/* Clamp long int to <0;100> */
static unsigned int normalize_log_level(long int raw) {
    if (raw < 0) {
        printf(_("Configured log level is less then 0. Clamped to 0.\n"));
        return 0;
    }
    if (raw > ILL_ALL) {
        printf(_("Configured log level is greater then %1$u. "
                    "Clamped to %1$u.\n"), ILL_ALL);
        return ILL_ALL;
    }
    if (raw > UINT_MAX ) {
        printf(_("Configured log level is greater then %1$u. "
                    "Clamped to %1$u.\n"), UINT_MAX);
        return UINT_MAX;
    }
    return (unsigned int) raw;
}


static int shi_init(const char *config_file) {
    isds_error err;
    char *value;
    unsigned int timeout, log_level;
    isds_log_facility log_facility = ILF_NONE;

    printf(_("This is Shigofumi, an ISDS client. Have a nice e-government.\n"));

    /* Do not permute arguments in getopt() */
    if (setenv("POSIXLY_CORRECT", "", 1)) {
        fprintf(stderr,
                _("Could not set POSIXLY_CORRECT environment variable\n"));
        return -1;
    }

    /* Load configuration */
    if (shi_load_configuration(config_file))
        return -1;
    timeout = normalize_timeout(cfg_getint(configuration, CONFIG_TIMEOUT));
    log_level = normalize_log_level(cfg_getint(configuration, CONFIG_LOGLEVEL));

    /* Init readline */
    rl_readline_name = "shigofumi";
    rl_filename_quote_characters = "\\ >";
    rl_filename_quoting_function = shi_quote_filename;
    rl_filename_dequoting_function = shi_dequote_filename;
    rl_char_is_quoted_p = shi_char_is_quoted;

    /* Initialize ISDS */ 
    err = isds_init();
    if (err) {
        fprintf(stderr, _("Could not initialize libisds library: %s\n"),
                isds_strerror(err));
        return -1;
    }

    /* Set ISDS logging */
    value = cfg_getstr(configuration, CONFIG_LOGFILE);
    if (do_log_to_file(value)) {
        fprintf(stderr, _("Could not redirect ISDS log to file `%s'\n"), value);
        return -1;
    }
    for (int i = 0; i < cfg_size(configuration, CONFIG_LOGFACILITIES); i++) {
        if (add_log_facility(&log_facility,
                cfg_getnstr(configuration, CONFIG_LOGFACILITIES, i)))
            return -1;

    }
    isds_set_logging(log_facility, log_level);

    /* Set ISDS context up */
    cisds = isds_ctx_create();
    if (!cisds) {
        fprintf(stderr, _("Could not create ISDS context\n"));
        return -1;
    }
    err = isds_set_timeout(cisds, timeout);
    if (err) {
        fprintf(stderr, _("Could not set ISDS network timeout: %s\n"),
             isds_strerror(err));
    }
    err = isds_set_progress_callback(cisds, shi_progressbar, NULL);
    if (err) {
        fprintf(stderr, _("Could not register network progress bar: %s: %s\n"),
                isds_strerror(err), isds_long_message(cisds));
    }
    err = isds_set_opt(cisds, IOPT_NORMALIZE_MIME_TYPE,
            cfg_getbool(configuration, CONFIG_NORMALIZEMIMETYPE));
    if (err) {
        fprintf(stderr,
                cfg_getbool(configuration, CONFIG_NORMALIZEMIMETYPE) ? 
                    _("Could not enable MIME type normalization: %s: %s\n") :
                    _("Could not enable MIME type normalization: %s: %s\n"),
                isds_strerror(err), isds_long_message(cisds));
    }
    if (!cfg_getbool(configuration, CONFIG_VERIFYSERVER)) {
        printf(_("Warning: Shigofumi disabled server identity verification "
                    "on user request!\n"));
        err = isds_set_opt(cisds, IOPT_TLS_VERIFY_SERVER, 0);
        if (err) {
            fprintf(stderr,
                    _("Could not disable server identity verification: "
                        "%s: %s\n"),
                    isds_strerror(err), isds_long_message(cisds));
        }
    }
    if ((value = cfg_getstr(configuration, CONFIG_CAFILE))) {
        err = isds_set_opt(cisds, IOPT_TLS_CA_FILE, value);
        if (err) {
            fprintf(stderr,
                    _("Could not set file with CA certificates: %s: %s: %s\n"),
                    value, isds_strerror(err), isds_long_message(cisds));
        }
    }
    if ((value = cfg_getstr(configuration, CONFIG_CADIRECTORY))) {
        err = isds_set_opt(cisds, IOPT_TLS_CA_DIRECTORY, value);
        if (err) {
            fprintf(stderr,
                    _("Could not set directory with CA certificates: "
                    "%s: %s: %s\n"),
                    value, isds_strerror(err), isds_long_message(cisds));
        }
    }
    if ((value = cfg_getstr(configuration, CONFIG_CRLFILE))) {
        err = isds_set_opt(cisds, IOPT_TLS_CRL_FILE, value);
        if (err) {
            fprintf(stderr, _("Could not set file with CRL: %s: %s: %s\n"),
                    value, isds_strerror(err), isds_long_message(cisds));
        }
    }
    

    /* Set Czech POINT context up */
    czechpoint = isds_ctx_create();
    if (!czechpoint) {
        fprintf(stderr, _("Could not create Czech POINT context\n"));
        return -1;
    }
    err = isds_set_timeout(czechpoint, timeout);
    if (err) {
        fprintf(stderr, _("Could not set Czech POINT network timeout: %s\n"),
             isds_strerror(err));
    }
    err = isds_set_progress_callback(czechpoint, shi_progressbar, NULL);
    if (err) {
        fprintf(stderr, "Could not register network progress bar: %s: %s\n",
                isds_strerror(err), isds_long_message(cisds));
    }

    return 0;
}


static int shi_quit(int argc, const char **argv) {
    shi_exit(EXIT_SUCCESS);
    return 0;
}



static void shi_help_usage(const char *command) {
    printf(_(
        "Usage: %s [COMMAND]\n"
        "Show COMMAND manual or list of currently available commands.\n"
        ),
            command);
}


static int shi_help(int argc, const char **argv) {
    size_t command_width = 14;
    int i;

    if (!commands) {
        printf(_("No command is available\n"));
        return 0;
    }

    if (argc == 2 && argv[1] && *argv[1]) {
        /* Show usage for given command */
        for (i = 0; (*commands)[i].name; i++) {
            if (!strcmp((*commands)[i].name, argv[1])) {
                if ((*commands)[i].usage)
                    (*commands)[i].usage((*commands)[i].name);
                else if ((*commands)[i].description) {
                    fhprint(stdout,  (*commands)[i].name, command_width);
                    printf("  %s\n", _((*commands)[i].description));
                }
                else
                    printf(_("%s: %s: Command description not defined\n"),
                            argv[0], argv[1]);
                return 0;
            }
        }
        printf(_("%s: %s: No such command exists\n"), argv[0], argv[1]);
        return -1;
    }

    /* Or list all commands */
    printf(_("Following commands are available:\n"));
    for (i = 0; (*commands)[i].name; i++) {
        fhprint(stdout,  (*commands)[i].name, command_width);
        printf("  %s\n", _((*commands)[i].description));
    }

    return 0;
}


static void show_version(void) {
    char *libisds_version = isds_version();

    printf(_("This is Shigofumi version %s.\n"), PACKAGE_VERSION);
    printf(_("\n"
                "Used libraries\n"
                "Readline: %s\n"
                "libisds: %s\n"
                ),
            rl_library_version, libisds_version);
    free(libisds_version);
}


static int shi_version(int argc, const char **argv) {
    show_version();
    printf(_(
"\n"
"-----\n"
"It's a shigofumi. A letter delivered from the afterlife.    (Fumika)\n"
"A message can not be delivered to dead person.  (ISDS specification)\n"
"Virtual and real world. They can be compatible.     (Program author)\n"
    ));
    return 0;
}


static int shi_copying(int argc, const char **argv) {
    printf(_(
    "This is Shigofumi, an ISDS client.\n"
    "Copyright (C) 2010  Petr Pisar\n"
    "\n"
    "This program is free software: you can redistribute it and/or modify\n"
    "it under the terms of the GNU General Public License as published by\n"
    "the Free Software Foundation, either version 3 of the License, or\n"
    "(at your option) any later version.\n"
    "\n"
    "This program is distributed in the hope that it will be useful,\n"
    "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
    "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
    "GNU General Public License for more details.\n"
    "\n"
    "You should have received a copy of the GNU General Public License\n"
    "along with this program.  If not, see <http://www.gnu.org/licenses/>.\n"
    ));
    return 0;
}


static int shi_cache(int argc, const char **argv) {
    const struct isds_list *item;
    size_t i;

    if (boxes) {
        for (item = boxes, i = 0; item; item = item->next, i++);
        printf(_(
                    "Cached box list: %zu\n"),
                i);
    }

    if (messages) {
        printf(_(
                    "Cached message list:\n"
                    "\tDirection: %s\n"
                    "\tMessages: %'lu\n"),
                (messages_are_outgoing) ? _("Outgoing") : _("Incoming"),
                total_messages);
    }

    if (message) {
        printf(_("Cached message: %s\n"),
                (message->envelope && message->envelope->dmID) ?
                message->envelope->dmID : _("<Unknown ID>"));
    }

    return 0;
}


static void shi_chdir_usage(const char *command) {
    printf(_(
                "Usage: %s [DIRECTORY]\n"
                "Change working directory to DIRECTORY.\n"
                "If no DIRECTORY is supplied, HOME directory will be used.\n"),
            command);
}


static int shi_chdir(int argc, const char **argv) {
    const char *directory = NULL;

    if (!argv || argc > 2) {
        shi_chdir_usage((argv) ? argv[0] : NULL);
        return -1;
    }

    if (argc == 2 && argv[1] && *argv[1])
        directory = argv[1];
    else {
        directory = getenv("HOME");
        if (!directory) {
            printf("Environment variable HOME does not exist\n");
            return -1;
        }
    }
    if (chdir(directory)) {
        printf(_("Could not change working directory: %s: %s\n"), directory,
                strerror(errno));
        return -1;
    }
    
    return 0;
}


static int shi_pwd(int argc, const char **argv) {
    char *buffer = NULL, *newbuffer;
    size_t length = 0;

    while (length += 1024) {
        newbuffer = realloc(buffer, length);
        if (!newbuffer) {
            printf(_("Error: Not enough memory\n"));
            free(buffer);
            return -1;
        }
        buffer = newbuffer;

        if (getcwd(buffer, length)) {
            printf("%s\n", buffer);
            free(buffer);
            return 0;
        }
    }

    printf(_("Error: Current directory string is too long\n"));
    free(buffer);
    return -1;
}


/* Log-in to ISDS */
static int do_login(void) {
    isds_error err;

    if (batch_mode) {
        printf(_("Unattended mode detected. "
                    "Make sure credentials have been preset.\n"));
    } else {
        printf(_("You are going to insert credentials for your account.\n"
                    "Leave blank line to choose default value.\n"));
    }
    select_completition(COMPL_NONE);

    /* Ask for server base URL */
    shi_replace_string(&server, _("Input ISDS base URL: "),
            cfg_getstr(configuration, CONFIG_SERVER), batch_mode);

    /* Ask for user name */
    shi_replace_string(&username, _("Input ISDS user name: "),
            cfg_getstr(configuration, CONFIG_USERNAME), batch_mode);

    /*if (!password) {
        password = ask_for_password(_("Input ISDS password: "));

        if (!password) {
            fprintf(stderr, _("Could not read ISDS password\n"));
            shi_exit(EXIT_FAILURE);
        }
    }*/
    shi_replace_password(&password, _("Input ISDS password: "),
            cfg_getstr(configuration, CONFIG_PASSWORD), batch_mode);

    select_completition(COMPL_COMMAND);
    set_prompt(NULL);

    printf(_("Logging in...\n"));
    err = isds_login(cisds, server, username, password, NULL);
    finish_isds_operation(cisds, err);
    if (err) {
        printf(_("Log-in failed\n"));
        return -1;
    }

    printf(_("Logged in.\n"));
    return 0;
}


static struct isds_DbOwnerInfo *do_box(void) {
    isds_error err;
    struct isds_DbOwnerInfo *box = NULL;

    printf(_("Getting box details you are logged in...\n"));
    err = isds_GetOwnerInfoFromLogin(cisds, &box);
    finish_isds_operation(cisds, err);

    return box;
}


static int shi_box(int argc, const char **argv) {
    struct isds_DbOwnerInfo *box = NULL;

    box = do_box();
    if (!box) return -1;

    format_DbOwnerInfo(box);

    isds_DbOwnerInfo_free(&box);
    return 0;
}


/* Get info about box with @id.
 * @id is UTF-8 encoded
 * Return NULL in case of error, otherwise box description that caller must
 * free. */
static struct isds_DbOwnerInfo *stat_box(const char *id) {
    isds_error err;
    struct isds_DbOwnerInfo criteria;
    struct isds_list *boxes = NULL, *item;
    struct isds_DbOwnerInfo *box = NULL;
    char *id_locale = NULL;

    if (!id || !*id) return NULL;

    id_locale = utf82locale(id);
    memset(&criteria, 0, sizeof(criteria));
    criteria.dbID = (char *) id;

    printf(_("Getting details about box with ID `%s'...\n"), id_locale);
    err = isds_FindDataBox(cisds, &criteria, &boxes);
    finish_isds_operation(cisds, err);
    if (err) goto leave;

    for(item = boxes; item; item = item->next) {
        if (!item->data) continue;

        if (item->next) {
            printf(_("Error: More boxes match ID `%s'\n"), id_locale);
            goto leave;
        }

        box = (struct isds_DbOwnerInfo *) item->data;
        item->data = NULL;
        break;
    }

leave:
    free(id_locale);
    isds_list_free(&boxes);

    return box;
}


static void shi_commercial_usage(const char *command) {
    printf(_(
"Usage: %s [-0|-1] [BOX_ID]\n"
"Manipulate commercial receiving box status.\n"
"  -O      switch off receiving of commercial messages\n"
"  -1      switch on receiving of commercial messages\n"
"  BOX_ID  affects box with ID BOX_ID; default is box you are logged in\n"
"If no option is given, show current commercial receiving status.\n"),
            command);
}


/* Manipulate commercial receiving box status */
static int shi_commercial(int argc, const char **argv) {
    isds_error err;
    struct isds_DbOwnerInfo *box = NULL;
    int opt;
    int action = -1;
    char *box_id = NULL, *box_id_locale = NULL;
    _Bool static_box_id;
    int retval = 0;

    optind = 0;
    while ((opt = getopt(argc, (char * const *)argv, "01")) != -1) {
        switch (opt) {
            case '0':
                action = 0;
                break;
            case '1':
                action = 1;
                break;
            default:
                shi_commercial_usage((argv)?argv[0]:NULL);
                return -1;
        }
    }
    if (optind + 1 < argc) {
        printf(_("Bad invocation\n"));
        shi_commercial_usage((argv)?argv[0]:NULL);
        return -1;
    }

    if (!argv[optind] || !*argv[optind]) {
        /* Get current box ID */
        box = do_box();
        if (!box || !box->dbID || !*box->dbID) {
            isds_DbOwnerInfo_free(&box);
            printf(_("Could not get current box ID\n"));
            return -1;
        }
        box_id = box->dbID; static_box_id = 1;
        box_id_locale = utf82locale(box_id);
    } else {
        /* Box ID supplied as argument */
        box_id_locale = (char *) argv[optind];
        box_id = locale2utf8(box_id_locale); static_box_id = 0;
        if (!box_id) {
            printf(_("Could not convert box ID `%s' to UTF-8\n"),
                    box_id_locale);
            return -1;
        }
    }

    if (action == -1) {
        if (!box) box = stat_box(box_id);
        if (!box) {
            printf(_("Could not get details about box ID `%s'\n"),
                    box_id_locale);
            retval = -1;
            goto leave;
        }

        printf(_("Commercial receiving status of box `%s': "), box_id_locale);
        if (!box->dbOpenAddressing)
            printf(_("Unknown\n"));
        else if (*box->dbOpenAddressing)
            printf(_("Positive\n"));
        else
            printf(_("Negative\n"));
    } else {
        char *refnumber = NULL;
        printf((action) ?
                _("Switching `%s' box commercial receiving on...\n"):
                _("Switching `%s' box commercial receiving off...\n"),
                box_id_locale);
        err = isds_switch_commercial_receiving(cisds, box_id, action,
                NULL, &refnumber);
        finish_isds_operation(cisds, err);

        if (!err) {
            char *refnumber_locale = utf82locale(refnumber);
            printf(_("Commercial receiving status successfully changed. "
                        "Assigned reference number: %s\n"),
                    refnumber_locale);
        } else {
            printf(_("Commercial receiving status has not been changed.\n"));
            retval = -1;
        }
        free(refnumber);
    }

leave:
    if (!static_box_id) free(box_id);
    else free(box_id_locale);
    isds_DbOwnerInfo_free(&box);
    return retval;
}


static int shi_user(int argc, const char **argv) {
    isds_error err;
    struct isds_DbUserInfo *user = NULL;

    printf(_("Getting user details you are logged as...\n"));
    err = isds_GetUserInfoFromLogin(cisds, &user);
    finish_isds_operation(cisds, err);
    if (err) return -1;

    format_DbUserInfo(user);

    isds_DbUserInfo_free(&user);
    return 0;
}


static void shi_users_usage(const char *command) {
    printf(_(
                "Usage: %s BOX_ID\n"
                "Get list of users having access to box with BOX_ID.\n"),
            command);
}


static int shi_users(int argc, const char **argv) {
    isds_error err;
    struct isds_list *users = NULL, *item;
    int ordinar;

    if (!argv || !argv[1] || !*argv[1]) {
        shi_users_usage((argv)?argv[0]:NULL);
        return -1;
    }

    printf(_("Getting users of box with ID `%s'...\n"), argv[1]);
    err = isds_GetDataBoxUsers(cisds, argv[1], &users);
    finish_isds_operation(cisds, err);
    if (err) return -1;

    for (item = users, ordinar = 0; item; item=item->next) {
        if (!item->data) continue;
        ordinar++;
        printf(_("\n* User #%d:\n"), ordinar);
        format_DbUserInfo(item->data);
    }
    if (ordinar == 0)
        printf(_("Empty list of users returned.\n"));

    isds_list_free(&users);
    return 0;
}


static int show_password_expiration(void) {
    isds_error err;
    struct timeval *expiration = NULL;

    err = isds_get_password_expiration(cisds, &expiration);
    finish_isds_operation(cisds, err);
    if (err) {
        fprintf(stderr, "Could not get password expiration time\n");
        return -1;
    }

    print_header_timeval(_("Your password expires at"), expiration);
    free(expiration);
    return 0;
}


/* Change password in ISDS */
static int do_passwd(void) {
    char *old_password = NULL;
    char *new_password = NULL;
    char *new_password2 = NULL;
    isds_error err = IE_ERROR;
    int retval = 0;

    select_completition(COMPL_NONE);

    printf(_(
"You are going to change your password. If you don't want to change your\n"
"password, insert empty string or EOF.\n"
"\n"
"You will be asked for your current (old) password and then for new password.\n"
"ISDS forces some criteria new password must fulfill. Current rules are:\n"
"\tLength: minimal 8, maximal 32 characters\n"
"\tMust contain at least: 1 upper case letter, 1 lower case letter, 1 digit\n"
"\tAllowed alphabet: [a-z][A-Z][0-9][!#$%%&()*+,-.:=?@[]_{}|~]\n"
"\tMust differ from last 255 passwords\n"
"\tMust not contain user ID\n"
"\tMust not contain sequence of three or more same characters\n"
"\tMust not start with `qwert', `asdgf', or `12345'\n"
"Finally, you must repeat your new password to avoid mistakes.\n"
"After password change will be confirmed, you must log in again as password\n"
"is transmitted to server on each request.\n"
"\n"));

    old_password = ask_for_password(_("Old password: "));
    if (!old_password || *old_password == '\0') {
        fprintf(stderr, _("No password supplied\n"));
        goto error;
    }

    new_password = ask_for_password(_("New password: "));
    if (!new_password || *new_password == '\0') {
        fprintf(stderr, _("No password supplied\n"));
        goto error;
    }

    new_password2 = ask_for_password(_("Repeat new password: "));
    if (!new_password2 || new_password2 == '\0') {
        fprintf(stderr, _("No password supplied\n"));
        goto error;
    }

    if (strcmp(new_password, new_password2)) {
        fprintf(stderr, _("New passwords differ\n"));
        goto error;
    }

    printf(_("Changing password...\n"));
    err = isds_change_password(cisds, old_password, new_password);
    finish_isds_operation(cisds, err);
    if (err) {
        printf(_("Password change failed\n"));
        goto error;
    } else {
        printf(_("Password HAS been successfully changed.\n"));
        goto leave;
    }

error:
    retval = -1;
    printf(_("Password has NOT been changed!\n"));

leave:
    free(old_password);
    free(new_password);
    free(new_password2);

    set_prompt(NULL);
    select_completition(COMPL_COMMAND);

    printf(_("\n"
            "Remember, ISDS password has limited life time.\n"));
    return retval;
}


static void shi_passwd_usage(const char *command) {
    printf(_(
"Usage: %s [-S]\n"
"Manipulate user password or change it if no option given.\n"
"\n"
"Options:\n"
"  -S     show password expiration time\n"
                ), command);
}


static int shi_passwd(int argc, const char **argv) {
    int opt;

    optind = 0;
    while ((opt = getopt(argc, (char * const *)argv, "S")) != -1) {
        switch (opt) {
            case 'S':
                return show_password_expiration();
            default:
                shi_passwd_usage(argv[0]);
                return -1;
        }
    }
    if (optind != argc || argc > 1) {
        printf(_("Bad invocation\n"));
        shi_passwd_usage((argv)?argv[0]:NULL);
        return -1;
    }

    return do_passwd();
}


static int shi_login(int argc, const char **argv) {
    discard_credentials();
    if (do_login()) return -1;
    show_password_expiration();
    return 0;
}


static void shi_debug_usage(const char *command) {
    printf(_(
"Usage: %s -l LEVEL [-f FACILITY...] [{-e | -o FILE}]\n"
"Debug FACILITIES on LEVEL.\n"
"\n"
"-l LEVEL       set log level, valid interval <%d,%d>, default is %d\n"
"               %d is no logging, %d critical, %d errors,\n"
"               %d warnings, %d info, %d debug, %d all\n"
"-f FACILITY    debug only given facility, repeat this option to debug\n"
"               more facilities; valid values: none, http, soap, isds,\n"
"               file, sec, xml, all; default is none\n"
"-e             write debug log into stderr\n"
"-o FILE        append debug log to FILE\n"
        ),
            command,
            ILL_NONE, ILL_ALL, ILL_NONE,
                ILL_NONE, ILL_CRIT, ILL_ERR, ILL_WARNING,
                ILL_INFO, ILL_DEBUG, ILL_ALL);
}

static int shi_debug(int argc, const char **argv) {
    int opt;
    int log_level = ILL_NONE;
    isds_log_facility log_facility = ILF_NONE;
    char *file = NULL;
    _Bool close_log = 0;

    optind = 0;
    while ((opt = getopt(argc, (char * const *)argv, "l:f:eo:")) != -1) {
        switch (opt) {
            case 'l':
                log_level = normalize_log_level(atoi(optarg));
                break;
            case 'f':
                if (add_log_facility(&log_facility, optarg)) return -1;
                break;
            case 'e':
                close_log = 1;
            case 'o':
                file = optarg;
                break;
            default:
                shi_debug_usage(argv[0]);
                return -1;
        }
    }
    if (optind == 1 || optind != argc) {
        printf(_("Bad invocation\n"));
        shi_debug_usage(argv[0]);
        return -1;
    }

    /* Redirect log */
    if (close_log) {
        isds_set_log_callback(NULL, NULL);

        if (logger_fd != -1) {
            if (-1 == close(logger_fd)) {
                fprintf(stderr, _("Closing log file failed: %s\n"),
                        strerror(errno));
                return -1;
            }
        }
    }
    if (do_log_to_file(file))
        return -1;
        
    /* Set log levels */
    isds_set_logging(log_facility, log_level);
    return 0;
}


static void shi_find_box_usage(const char *command) {
    printf(_(
"Usage: %s {OPTION... | BOX_ID}\n"
"Get information about box with BOX_ID or boxes meeting other criteria.\n"
"Each search option requires an argument:\n"
"  -t     box type; accepted values:\n"
"             FO            Private individual\n"
"             PFO           Self-employed individual\n"
"             PFO_ADVOK     Lawyer\n"
"             PFO_DANPOR    Tax advisor\n"
"             PFO_INSSPR    Insolvency administrator\n"
"             PO            Organisation\n"
"             PO_ZAK        Organization based by law\n"
"             PO_REQ        Organization based on request\n"
"             OVM           Public authority\n"
"             OVM_NOTAR     Notary\n"
"             OVM_EXEKUT    Executor\n"
"             OVM_REQ       Public authority based on request\n"
"  -j     identity number\n"
"\n"
"Person name options:\n"
"  -f     first name\n"
"  -m     middle name\n"
"  -l     last name\n"
"  -b     last name at birth\n"
"  -s     subject name\n"
"\n"
"Birth options:\n"
"  -d     birth date (locale or full ISO 8601 date)\n"
"  -w     birth city\n"
"  -y     birth county\n"
"  -c     birth state\n"
"\n"
"Address:\n"
"  -W     city\n"
"  -S     street\n"
"  -z     number in street\n"
"  -Z     number in municipality\n"
"  -P     ZIP code\n"
"  -C     state\n"
"\n"
"Other options:\n"
"  -n     nationality\n"
"  -e     e-mail\n"
"  -p     phone number\n"
"  -i     identifier\n"
"  -r     registry code\n"
"  -a     box status; accepted values:\n"
"             ACCESSIBLE            Accessible\n"
"             TEMP_INACCESSIBLE     Temporary inaccessible\n"
"             NOT_YET_ACCESSIBLE    Not yet accessible\n"
"             PERM_INACCESSIBLE     Permanently inaccessible\n"
"             REMOVED               Deleted\n"
"  -o     act as public authority; boolean values: 0 is false, 1 is true\n"
"  -k     receive commercial messages; boolean values\n"
"\n"
"Not all option combinations are meaningful or allowed. For example box\n"
"type is always required (except direct box ID query).\n"
"ISDS can refuse to answer to much broad query. Not all boxes are searchable\n"
"by every user.\n"
),
            command);
}


/* Allow  reassignment */
#define FILL_OR_LEAVE(variable, locale) { \
    zfree(variable); \
    (variable) = locale2utf8(locale); \
    if (!(variable)) { \
        printf(_("Error: Not enough memory\n")); \
        retval = -1; \
        goto leave; \
    } \
}

#define CALLOC_OR_LEAVE(structure) { \
    if (!(structure)) { \
        (structure) = calloc(1, sizeof(*(structure))); \
        if (!(structure)) { \
            printf(_("Error: Not enough memory\n")); \
            retval = -1; \
            goto leave; \
        } \
    } \
}

#define FILL_BOOLEAN_OR_LEAVE(variable, locale) { \
    zfree(variable); \
    (variable) = malloc(sizeof(*(variable))); \
    if (!(variable)) { \
        printf(_("Error: Not enough memory\n")); \
        retval = -1; \
        goto leave; \
    } \
    if (!strcmp((locale), "0")) *(variable) = 0; \
    else if (!strcmp((locale), "1")) *(variable) = 1; \
    else { \
        printf(_("%s: %s: Unknown boolean value\n"), argv[0], (locale)); \
        retval = -1; \
        goto leave; \
    } \
}

#define FILL_LONGINT_OR_LEAVE(variable, locale) { \
    if (!(locale) || !*(locale)) { \
        printf(_("%s: Empty integer value\n"), argv[0]); \
        retval = -1; \
        goto leave; \
    } \
    char *endptr; \
    zfree(variable); \
    (variable) = malloc(sizeof(*(variable))); \
    if (!(variable)) { \
        printf(_("Error: Not enough memory\n")); \
        retval = -1; \
        goto leave; \
    } \
    (*variable) = strtol((locale), &endptr, 0); \
    if (*endptr) { \
        printf(_("%s: %s: Invalid integer value\n"), argv[0], (locale)); \
        retval = -1; \
        goto leave; \
    } \
}

static int shi_find_box(int argc, const char **argv) {
    int opt;
    isds_error err;
    struct isds_DbOwnerInfo *criteria = NULL;
    struct isds_list *item;
    int order = 0;
    int retval = 0;

    if (!argv || !argv[1] || !*argv[1]) {
        printf(_("Error: No argument supplied\n"));
        shi_find_box_usage((argv)?argv[0]:NULL);
        return -1;
    }

    criteria = calloc(1, sizeof(*criteria));
    if (!criteria) {
        printf(_("Error: Not enough memory\n"));
        retval = -1;
        goto leave;
    }

    /* Parse options */
    optind = 0;
    while ((opt = getopt(argc, (char * const *)argv, "t:j:s:"
                    "f:m:l:b:s:" "d:w:y:c:" "W:S:z:Z:P:C:"
                    "n:e:p:i:r:a:o:k:")) != -1) {
        switch (opt) {
            case 't':
                criteria->dbType = malloc(sizeof(*criteria->dbType));
                if (!criteria->dbType) {
                    printf(_("Error: Not enough memory\n"));
                    retval = -1;
                    goto leave;
                }
                if (!strcmp(optarg, "FO"))
                    *criteria->dbType = DBTYPE_FO;
                else if (!strcmp(optarg, "PFO"))
                    *criteria->dbType = DBTYPE_PFO;
                else if (!strcmp(optarg, "PFO_ADVOK"))
                    *criteria->dbType = DBTYPE_PFO_ADVOK;
                else if (!strcmp(optarg, "PFO_DANPOR"))
                    *criteria->dbType = DBTYPE_PFO_DANPOR;
                else if (!strcmp(optarg, "PFO_INSSPR"))
                    *criteria->dbType = DBTYPE_PFO_INSSPR;
                else if (!strcmp(optarg, "PO"))
                    *criteria->dbType = DBTYPE_PO;
                else if (!strcmp(optarg, "PO_ZAK"))
                    *criteria->dbType = DBTYPE_PO_ZAK;
                else if (!strcmp(optarg, "PO_REQ"))
                    *criteria->dbType = DBTYPE_PO_REQ;
                else if (!strcmp(optarg, "OVM"))
                    *criteria->dbType = DBTYPE_OVM;
                else if (!strcmp(optarg, "OVM_NOTAR"))
                    *criteria->dbType = DBTYPE_OVM_NOTAR;
                else if (!strcmp(optarg, "OVM_EXEKUT"))
                    *criteria->dbType = DBTYPE_OVM_EXEKUT;
                else if (!strcmp(optarg, "OVM_REQ"))
                    *criteria->dbType = DBTYPE_OVM_REQ;
                else {
                    printf(_("%s: %s: Unknown box type\n"), argv[0], optarg);
                    retval = -1;
                    goto leave;
                }
                break;

            case 'j':
                FILL_OR_LEAVE(criteria->ic, optarg);
                break;

            /* Person name */
            case 'f':
                CALLOC_OR_LEAVE(criteria->personName);
                FILL_OR_LEAVE(criteria->personName->pnFirstName, optarg);
                break;
            case 'm':
                CALLOC_OR_LEAVE(criteria->personName);
                FILL_OR_LEAVE(criteria->personName->pnMiddleName, optarg);
                break;
            case 'l':
                CALLOC_OR_LEAVE(criteria->personName);
                FILL_OR_LEAVE(criteria->personName->pnLastName, optarg);
                break;
            case 'b':
                CALLOC_OR_LEAVE(criteria->personName);
                FILL_OR_LEAVE(criteria->personName->pnLastNameAtBirth, optarg);
                break;
            case 's':
                FILL_OR_LEAVE(criteria->firmName, optarg);
                break;

            /* Birth */
            case 'd':
                CALLOC_OR_LEAVE(criteria->birthInfo);
                criteria->birthInfo->biDate = datestring2tm(optarg);
                if (!criteria->birthInfo->biDate) {
                    printf(_("Error: Could not parse date: %s\n"), optarg);
                    retval = -1;
                    goto leave;
                }
                break;
            case 'w':
                CALLOC_OR_LEAVE(criteria->birthInfo);
                FILL_OR_LEAVE(criteria->birthInfo->biCity, optarg);
                break;
            case 'y':
                CALLOC_OR_LEAVE(criteria->birthInfo);
                FILL_OR_LEAVE(criteria->birthInfo->biCounty, optarg);
                break;
            case 'c':
                CALLOC_OR_LEAVE(criteria->birthInfo);
                FILL_OR_LEAVE(criteria->birthInfo->biState, optarg);
                break;

            /* Address */
            case 'W':
                CALLOC_OR_LEAVE(criteria->address);
                FILL_OR_LEAVE(criteria->address->adCity, optarg);
                break;
            case 'S':
                CALLOC_OR_LEAVE(criteria->address);
                FILL_OR_LEAVE(criteria->address->adStreet, optarg);
                break;
            case 'z':
                CALLOC_OR_LEAVE(criteria->address);
                FILL_OR_LEAVE(criteria->address->adNumberInStreet, optarg);
                break;
            case 'Z':
                CALLOC_OR_LEAVE(criteria->address);
                FILL_OR_LEAVE(criteria->address->adNumberInMunicipality,
                        optarg);
                break;
            case 'P':
                CALLOC_OR_LEAVE(criteria->address);
                FILL_OR_LEAVE(criteria->address->adZipCode, optarg);
                break;
            case 'C':
                CALLOC_OR_LEAVE(criteria->address);
                FILL_OR_LEAVE(criteria->address->adState, optarg);
                break;

            /* Other options */
            case 'n':
                FILL_OR_LEAVE(criteria->nationality, optarg);
                break;
            case 'e':
                FILL_OR_LEAVE(criteria->email, optarg);
                break;
            case 'p':
                FILL_OR_LEAVE(criteria->telNumber, optarg);
                break;
            case 'i':
                FILL_OR_LEAVE(criteria->identifier, optarg);
                break;
            case 'r':
                FILL_OR_LEAVE(criteria->registryCode, optarg);
                break;
            case 'a':
                criteria->dbState = malloc(sizeof(*criteria->dbState));
                if (!criteria->dbState) {
                    printf(_("Error: Not enough memory\n"));
                    retval = -1;
                    goto leave;
                }
                if (!strcmp(optarg, "ACCESSIBLE"))
                    *criteria->dbState = DBSTATE_ACCESSIBLE;
                else if (!strcmp(optarg, "TEMP_INACCESSIBLE"))
                    *criteria->dbState = DBSTATE_TEMP_UNACCESSIBLE;
                else if (!strcmp(optarg, "NOT_YET_ACCESSIBLE"))
                    *criteria->dbState = DBSTATE_NOT_YET_ACCESSIBLE;
                else if (!strcmp(optarg, "PERM_INACCESSIBLE"))
                    *criteria->dbState = DBSTATE_PERM_UNACCESSIBLE;
                else if (!strcmp(optarg, "REMOVED"))
                    *criteria->dbState = DBSTATE_REMOVED;
                else {
                    printf(_("%s: %s: Unknown box status\n"), argv[0], optarg);
                    retval = -1;
                    goto leave;
                }
                break;
            case 'o':
                FILL_BOOLEAN_OR_LEAVE(criteria->dbEffectiveOVM, optarg);
                break;
            case 'k':
                FILL_BOOLEAN_OR_LEAVE(criteria->dbOpenAddressing, optarg);
                break;

            default:
                shi_find_box_usage(argv[0]);
                retval = -1;
                goto leave;
        }
    }

    /* There must be an option and all of them must be recognized, if not only
     * BOX_ID supplied */
    if (argc > 2 && optind != argc) {
        printf(_("Error: Superfluous argument\n"));
        shi_find_box_usage(argv[0]);
        retval = -1;
        goto leave;
    }
   
    /* If only box ID is supplied use it */
    if (argc == 2 && argv[1] && *argv[1]) {
        criteria->dbID = locale2utf8(argv[1]);
        if (!criteria->dbID) {
            printf(_("Error: Not enough memory\n"));
            retval = -1;
            goto leave;
        }
    }

    printf(_("Searching boxes...\n"));
    err = isds_FindDataBox(cisds, criteria, &boxes);
    finish_isds_operation(cisds, err);
    if (err) return -1;

    for(item = boxes; item; item = item->next) {
        if (!item->data) continue;
        order++;
        
        printf(_("\n* Result #%d:\n"), order);
        format_DbOwnerInfo(item->data);
    }

leave:
    isds_DbOwnerInfo_free(&criteria);
    return retval;
}


static void shi_stat_box_usage(const char *command) {
    printf(_(
"Usage: %s BOX_ID...\n"
"Get status of box with BOX_ID. More boxes can be specified.\n"),
            command);
}


/* Get boxes status */
static int shi_stat_box(int argc, const char **argv) {
    isds_error err;
    char *id = NULL;
    long int status;

    if (!argv || !*argv || argc < 2 || !argv[1]) {
        printf(_("Missing box ID\n"));
        shi_stat_box_usage((argv[0])?argv[0]:NULL);
        return -1;
    }

    for (int i = 1; i < argc; i++) {
        if (!argv[i] || !*argv[i]) continue;

        free(id);
        id = locale2utf8(argv[i]);
        if (!id) {
            printf(_("%s: Could not covert box ID to UTF-8\n"), argv[i]);
            return -1;
        }

        printf(_("Getting status of box `%s'...\n"), argv[i]);
        err = isds_CheckDataBox(cisds, id, &status);
        finish_isds_operation(cisds, err);
        if (err) return -1;
        
        printf(_("Status of box `%s': %s\n"),
                argv[i], DbState2string(&status));
    }

    return 0;
}


static void shi_compose_usage(const char *command) {
    printf(_(
"Usage: %s OPTION...\n"
"Compose and send a message to recipient defined by his box ID.\n"
"Each option requires an argument (if not stated otherwise):\n"
"  -s *   message subject\n"
"\n"
"Recipient options:\n"
"  -b *   recipient box ID\n"
"  -U     organisation unit name\n"
"  -N     organisation unit number\n"
"  -P     to hands of given person\n"
"\n"
"Sender organisation structure options:\n"
"  -u     unit name\n"
"  -n     unit number\n"
"\n"
"Message identifier options:\n"
"  -r     sender reference number\n"
"  -f     sender file ID\n"
"  -R     recipient reference number\n"
"  -F     recipient file ID\n"
"\n"
"Legal title options:\n"
"  -y     year act has been issued\n"
"  -a     ordinal number of act in a year\n"
"  -e     section of the act\n"
"  -o     paragraph of the act\n"
"  -i     point of the paragraph of the act\n"
"\n"
"Delivery options:\n"
"  -p     personal delivery required\n"
"  -t     allow substitutable delivery\n"
"  -A     non-OVM sender acts as public authority\n"
"\n"
"Document options:\n"
"  -d *   read document from local file\n"
"  -D     document name (defaults to base local file name)\n"
"  -x     transport subset of the document as a XML.\n"
"         Argument is XPath expression specifying desired node set\n"         
"  -m     override MIME type (guessed on -d)\n"
"  -g     document ID (must be unique per message)\n"
"  -G     reference to other document using its ID\n"
"  -c     document is digital signature of other document (NO argument\n"
"         allowed)\n"
"\n"
"Options marked with asterisk are mandatory, other are optional. Another soft\n"
"dependencies can emerge upon using specific option. They are not mandated by\n"
"ISDS currently, but client library or this program can force them to assure\n"
"semantically complete message. Following soft dependencies are recommended:\n"
"  -y <=> -a  act number and year must be used at the same time\n"
"  -i => -o   act point requires act paragraph\n"
"  -o => -e   act paragraph requires act section\n"
"  -e => -a   act section requires act number\n"
"  -G => -g   document with referenced ID must exist\n"
"  -c => -G   signature must refer to signed document\n"
"  -c         first document cannot be signature\n"
"\n"
"More documents can be attached to a message by repeating `-d' option.\n"
"Document order will be preserved. Other document options affect immediately\n"
"preceding `-d' document only. E.g. `-d /tmp/foo.pdf -m application/pdf\n"
"-d /tmp/bar.txt -m text/plain' attaches first PDF file, then textual file.\n"
"\n"
"The same applies to recipient options that must start with box ID (-b).\n"
"If more recipients specified, each of them will get a copy of composed\n"
"message. ISDS will assign message identifier to each copy in turn.\n"
),
            command);
}


static int shi_compose(int argc, const char **argv) {
    int opt;
    isds_error err;
    int retval = 0;
    struct isds_message *message = NULL;
    struct isds_envelope *envelope = NULL;
    struct isds_list *documents = NULL;
    struct isds_document *document = NULL;
    struct isds_list *copies = NULL, *copy_item = NULL;
    struct isds_message_copy *copy = NULL;
    char *message_id_locale = NULL, *recipient_id_locale = NULL,
         *dmStatus_locale = NULL;

    if (!argv || !argv[1] || !*argv[1]) {
        printf(_("Error: No argument supplied\n"));
        shi_compose_usage((argv)?argv[0]:NULL);
        return -1;
    }

    message = calloc(1, sizeof(*message));
    if (!message) {
        printf(_("Error: Not enough memory\n"));
        retval = -1;
        goto leave;
    }
    envelope = calloc(1, sizeof(*envelope));
    if (!envelope) {
        printf(_("Error: Not enough memory\n"));
        retval = -1;
        goto leave;
    }
    message->envelope = envelope;

    /* Parse options */
    optind = 0;
    while ((opt = getopt(argc, (char * const *)argv, "s:" "b:U:N:P:" "u:n:"
                    "r:f:R:F:" "y:a:e:o:i:" "p:t:A:" "d:D:x:m:g:G:c"
                    )) != -1) {
        switch (opt) {
            case 's':
                FILL_OR_LEAVE(envelope->dmAnnotation, optarg);
                break;

            /* Recipient options */
            case 'b':
                copy = NULL;
                if (!copies) {
                    /* First recipient */
                    CALLOC_OR_LEAVE(copies);
                    copies->destructor =
                        (void(*)(void **)) isds_message_copy_free;
                    copy_item = copies;
                } else {
                    /* Next recipient */
                    CALLOC_OR_LEAVE(copy_item->next);
                    copy_item->next->destructor =
                        (void(*)(void **)) isds_message_copy_free;
                    copy_item = copy_item->next;
                }
                CALLOC_OR_LEAVE(copy);
                copy_item->data = copy;

                /* Copy recipient box ID */
                FILL_OR_LEAVE(copy->dbIDRecipient, optarg);
                break;
            case 'U':
                if (!copy) {
                    printf(_("Error: %s: Recipient box ID (-b) must precede "
                                "recipient organisation unit name (-%c)\n"),
                            optarg, opt);
                    retval = -1;
                    goto leave; 
                }
                FILL_OR_LEAVE(copy->dmRecipientOrgUnit, optarg);
                break;
            case 'N':
                if (!copy) {
                    printf(_("Error: %s: Recipient box ID (-b) must precede "
                                "recipient organisation unit number (-%c)\n"),
                            optarg, opt);
                    retval = -1;
                    goto leave; 
                }
                FILL_LONGINT_OR_LEAVE(copy->dmRecipientOrgUnitNum, optarg);
                break;
            case 'P':
                if (!copy) {
                    printf(_("Error: %s: Recipient box ID (-b) must precede "
                                "to-hands option (-%c)\n"), optarg, opt);
                    retval = -1;
                    goto leave; 
                }
                FILL_OR_LEAVE(copy->dmToHands, optarg);
                break;

            /* Sender organisation structure options */
            case 'u':
                FILL_OR_LEAVE(envelope->dmSenderOrgUnit, optarg);
                break;
            case 'n':
                FILL_LONGINT_OR_LEAVE(envelope->dmSenderOrgUnitNum, optarg);
                break;

            /* Message identifier options */
            case 'r':
                FILL_OR_LEAVE(envelope->dmSenderRefNumber, optarg);
                break;
            case 'f':
                FILL_OR_LEAVE(envelope->dmSenderIdent, optarg);
                break;
            case 'R':
                FILL_OR_LEAVE(envelope->dmRecipientRefNumber, optarg);
                break;
            case 'F':
                FILL_OR_LEAVE(envelope->dmRecipientIdent, optarg);
                break;

            /* Legal title options */
            case 'y':
                FILL_LONGINT_OR_LEAVE(envelope->dmLegalTitleYear, optarg);
                break;
            case 'a':
                FILL_LONGINT_OR_LEAVE(envelope->dmLegalTitleLaw, optarg);
                break;
            case 'e':
                FILL_OR_LEAVE(envelope->dmLegalTitleSect, optarg);
                break;
            case 'o':
                FILL_OR_LEAVE(envelope->dmLegalTitlePar, optarg);
                break;
            case 'i':
                FILL_OR_LEAVE(envelope->dmLegalTitlePoint, optarg);
                break;

            /* Delivery options */
            case 'p':
                FILL_BOOLEAN_OR_LEAVE(envelope->dmPersonalDelivery, optarg);
                break;
            case 't':
                FILL_BOOLEAN_OR_LEAVE(envelope->dmAllowSubstDelivery, optarg);
                break;
            case 'A':
                FILL_BOOLEAN_OR_LEAVE(envelope->dmOVM, optarg);
                break;

            /* Document options */
            case 'd':
                document = NULL;
                if (!documents) {
                    /* First document */
                    CALLOC_OR_LEAVE(message->documents);
                    message->documents->destructor =
                        (void(*)(void **)) isds_document_free;
                    documents = message->documents;
                    CALLOC_OR_LEAVE(document);
                    documents->data = document;
                    document->dmFileMetaType = FILEMETATYPE_MAIN;
                } else {
                    /* Next document */
                    CALLOC_OR_LEAVE(documents->next);
                    documents->next->destructor =
                        (void(*)(void **)) isds_document_free;
                    documents = documents->next;
                    CALLOC_OR_LEAVE(document);
                    documents->data = document;
                    document->dmFileMetaType = FILEMETATYPE_ENCLOSURE;
                }

                if (load_data_from_file(optarg, &document->data,
                        &document->data_length, &document->dmMimeType)) {
                    retval = -1;
                    goto leave;
                }
                /* XXX: POSIX basename() modifies argument */
                FILL_OR_LEAVE(document->dmFileDescr, basename(optarg));
                break;
            case 'D':
                if (!document) {
                    printf(_("Error: %s: Document file (-d) must precede "
                                "document name (-%c)\n"), optarg, opt);
                    retval = -1;
                    goto leave; 
                }
                FILL_OR_LEAVE(document->dmFileDescr, optarg);
                break;
            case 'x':
                if (!document) {
                    printf(_("Error: %s: Document file (-d) must precede "
                                "XPath expression (-%c)\n"), optarg, opt);
                    retval = -1;
                    goto leave; 
                }
                /* Load XML node list */
                char *xpath_expr = NULL;
                FILL_OR_LEAVE(xpath_expr, optarg);
                retval = load_xml_subtree_from_memory(
                        document->data, document->data_length,
                        &document->xml_node_list, xpath_expr);
                if (retval) {
                    free(xpath_expr);
                    goto leave;
                }
                /* Switch document type to XML */
                document->is_xml = 1;
                zfree(document->data);
                document->data_length = 0;
                documents->destructor =
                    (void(*)(void **)) free_document_with_xml_node_list;
                break;
            case 'm':
                if (!document) {
                    printf(_("Error: %s: Document file (-d) must precede "
                                "MIME type (-%c)\n"), optarg, opt);
                    retval = -1;
                    goto leave; 
                }
                FILL_OR_LEAVE(document->dmMimeType, optarg);
                break;
            case 'g':
                if (!document) {
                    printf(_("Error: %s: Document file (-d) must precede "
                                "document ID (-%c)\n"), optarg, opt);
                    retval = -1;
                    goto leave; 
                }
                FILL_OR_LEAVE(document->dmFileGuid, optarg);
                break;
            case 'G':
                if (!document) {
                    printf(_("Error: %s: Document file (-d) must precede "
                                "document reference (-%c)\n"), optarg, opt);
                    retval = -1;
                    goto leave; 
                }
                FILL_OR_LEAVE(document->dmUpFileGuid, optarg);
                break;
            case 'c':
                if (!document) {
                    printf(_("Error: Document file (-d) must precede "
                                "document signature type (-%c)\n"), opt);
                    retval = -1;
                    goto leave; 
                }
                document->dmFileMetaType = FILEMETATYPE_SIGNATURE;
                break;

            default:
                shi_compose_usage(argv[0]);
                retval = -1;
                goto leave;
        }
    }

    /* All options must be recognized */
    if (optind != argc) {
        printf(_("Error: Superfluous argument\n"));
        shi_compose_usage(argv[0]);
        retval = -1;
        goto leave;
    }

    if (!copies) {
        printf(_("Error: No recipient box ID specified\n"));
        shi_compose_usage(argv[0]);
        retval = -1;
        goto leave;
    }
   
    /* TODO: Check Legal Title soft dependencies */

    /* Preview */
    printf(_("Following message has been composed:\n"));
    format_message(message);
    printf(_("Following recipients have been specified:\n"));
    format_copies(copies);
    /* TODO: Confirmation, correction  */

    /* Send a message */
    printf(_("Sending message...\n"));
    err = isds_send_message_to_multiple_recipients(cisds, message, copies);
    finish_isds_operation(cisds, err);
    if (err && err != IE_PARTIAL_SUCCESS) {
        retval = -1;
        goto leave;
    }

    /* Show results for each copy */
    for (copy_item = copies; copy_item; copy_item = copy_item->next) {
        if (!copy_item->data) continue;
        copy = (struct isds_message_copy *) copy_item->data;
        recipient_id_locale = utf82locale(copy->dbIDRecipient);

        if (copy->error) {
            retval = -1;
            if (copy->dmStatus) dmStatus_locale = utf82locale(copy->dmStatus);
            if (dmStatus_locale)
                printf(_("%s: Failed: %s: %s\n"),
                        recipient_id_locale,
                        isds_strerror(copy->error), dmStatus_locale);
            else
                printf(_("%s: Failed: %s\n"), recipient_id_locale,
                        isds_strerror(copy->error));
            zfree(dmStatus_locale);
        } else {
            message_id_locale = utf82locale(copy->dmID);
            printf(_("%s: Succeeded. Assigned message ID: %s\n"),
                    recipient_id_locale, message_id_locale);
            free(message_id_locale);
        }

        free(recipient_id_locale);
    }

leave:
    isds_message_free(&message);
    isds_list_free(&copies);
    return retval;
}

#undef FILL_LONGINT_OR_LEAVE
#undef FILL_BOOLEAN_OR_LEAVE
#undef CALLOC_OR_LEAVE
#undef FILL_OR_LEAVE


static void shi_delivery_usage(const char *command) {
    printf(_(
                "Usage: %s MESSAGE_ID\n"
                "Get delivery data about message with MESSAGE_ID.\n"),
            command);
}


static int shi_delivery(int argc, const char **argv) {
    isds_error err;
    const char *id;

    if (!argv || !argv[1] || !*argv[1]) {
        shi_delivery_usage(argv[0]);
        return -1;
    }
    id = argv[1];

    printf(_("Getting delivery info...\n"));
    err = isds_get_signed_delivery_info(cisds, id, &message);
    finish_isds_operation(cisds, err);
    if (err) {
        set_prompt(NULL);
        select_completition(COMPL_COMMAND);
        return -1;
    }

    format_message(message);

    if (message->envelope && message->envelope->dmID)
        set_prompt(_("%s %s"), argv[0], message->envelope->dmID);
    else 
        set_prompt("%s", argv[0]);
    select_completition(COMPL_MSG);
    return 0;
}


static int shi_dump_message(int argc, const char **argv) {
    if (!message) {
        printf(_("No message loaded\n"));
        return -1;
    }

    print_message(message);
    return 0;
}


static void shi_hash_usage(const char *command) {
    printf(_(
        "Usage: %s [MESSAGE_ID]\n"
        "Retrieve message hash stored in ISDS.\n"
        "If MESSAGE_ID is defined, query for that message.\n"
        "Otherwise use current message.\n"),
            command);
}


static int shi_hash(int argc, const char **argv) {
    isds_error err;
    const char *id = NULL;
    struct isds_hash *hash = NULL;
    char *hash_string = NULL;

    if (!argv || argc > 2) {
        shi_hash_usage((argv)?argv[0]:NULL);
        return -1;
    }
    if (argc == 2 && argv[1] && *argv[1])
        id = argv[1];
    else {
        if (!message) {
            printf(_("No message loaded\n"));
            return -1;
        }
        if (!message->envelope || !message->envelope->dmID) {
            printf(_("Current message is missing ID\n"));
            return -1;
        }
        id = message->envelope->dmID;
    }

    printf(_("Getting message hash...\n"));
    err = isds_download_message_hash(cisds, id, &hash);
    finish_isds_operation(cisds, err);
    if (err) return -1;

    hash_string = hash2string(hash);
    printf(_("ISDS states message with `%s' ID has following hash:\n%s\n"),
            id, hash_string);

    free(hash_string);
    isds_hash_free(&hash);
    return 0;
}


static int shi_verify(int argc, const char **argv) {
    isds_error err;
    int retval = 0;
    struct isds_hash *retrieved_hash = NULL, *stored_hash = NULL;
    char *hash_string = NULL;
    size_t width = 4;

    if (!message) {
        printf(_("No message loaded\n"));
        return -1;
    }
  
    if (!message->envelope) {
        printf(_("Current message is missing envelope\n"));
        return -1;
    }
    stored_hash = message->envelope->hash;
    message->envelope->hash = NULL;
    
    if (message->envelope->dmID) {
        /* Verify remote hash */
        printf(_("Remote hash check:\n"));

        printf(_("Getting message hash...\n"));
        err = isds_download_message_hash(cisds, message->envelope->dmID,
                &retrieved_hash);
        finish_isds_operation(cisds, err);

        if (retrieved_hash) {
            hash_string = hash2string(retrieved_hash);
            fhprint(stdout, _("Retrieved:"), width);
            printf("%s\n", hash_string);
            zfree(hash_string);
        }

        if (retrieved_hash && message->raw) {
            err = isds_compute_message_hash(cisds, message,
                    retrieved_hash->algorithm);
            finish_isds_operation(cisds, err);

            if (!err) {
                hash_string = hash2string(message->envelope->hash);
                fhprint(stdout, _("Computed:"), width);
                printf("%s\n", hash_string);
                zfree(hash_string);
            }
        }

        err = isds_hash_cmp(retrieved_hash, message->envelope->hash);
        switch (err) {
            case IE_SUCCESS:
                printf(_("Hashes match.\n")); break;
            case IE_NOTUNIQ:
                printf(_("Hashes do not match.\n"));
                retval = -1;
                break;
            default:
                printf(_("Hashes could not be compared.\n"));
                retval = -1;
                break;
        }

        free(retrieved_hash);
    }


    if (stored_hash) {
        /* Verify stored hash */
        printf(_("Stored hash check:\n"));

        hash_string = hash2string(stored_hash);
        fhprint(stdout, _("Stored:"), width);
        printf("%s\n", hash_string);
        zfree(hash_string);

        if (message->raw) {
            err = isds_compute_message_hash(cisds, message,
                    stored_hash->algorithm);
            finish_isds_operation(cisds, err);

            if (!err) {
                hash_string = hash2string(message->envelope->hash);
                fhprint(stdout, _("Computed:"), width);
                printf("%s\n", hash_string);
                zfree(hash_string);
            }
        }

        err = isds_hash_cmp(stored_hash, message->envelope->hash);
        switch (err) {
            case IE_SUCCESS:
                printf(_("Hashes match.\n")); break;
            case IE_NOTUNIQ:
                printf(_("Hashes do not match.\n"));
                retval = -1;
                break;
            default:
                printf(_("Hashes could not be compared.\n"));
                retval = -1;
                break;
        }

        isds_hash_free(&message->envelope->hash);
    }

    message->envelope->hash = stored_hash;
    return retval;
}


static int shi_authenticate(int argc, const char **argv) {
    isds_error err;
    int retval = 0;

    if (!message) {
        printf(_("No message loaded\n"));
        return -1;
    }
    if (!message->raw || message->raw_length == 0) {
        printf(_("Current message is missing raw representation\n"));
        return -1;
    }

    printf(_("Submitting message to authenticity check...\n"));
    err = isds_authenticate_message(cisds, message->raw, message->raw_length);
    finish_isds_operation(cisds, (err == IE_NOTUNIQ) ? IE_SUCCESS : err);

    switch (err) {
        case IE_SUCCESS:
            printf(_("Message originates in ISDS.\n")); break;
        case IE_NOTUNIQ:
            printf(_("Message is unknown to ISDS or has been tampered.\n"));
            retval = -1;
            break;
        default:
            retval = -1;
            break;
    }

    return retval;
}


static void shi_accept_message_usage(const char *command) {
    printf(_(
"Usage: %s [MESSAGE_ID...]\n"
"Accept commercial message moving its state to received.\n"
"If MESSAGE_ID is defined, accept that message. More messages can be specified.\n"
"Otherwise accept all commercial incoming messages.\n"),
            command);
}


static int shi_accept_message(int argc, const char **argv) {
    isds_error err;
    char *id = NULL;

    /* Process messages named in argv */
    for (int i = 1; i < argc; i++) {
        if (!argv[i] || !*argv[i]) continue;

        id = locale2utf8(argv[i]);
        if (!id) {
            printf(_("Error: Could not convert message ID to UTF-8: %s\n"),
                    argv[i]);
            return -1;
        }

        printf(_("Accepting message `%s'...\n"), argv[i]);
        err = isds_mark_message_received(cisds, id);
        finish_isds_operation(cisds, err);
        if (err) {
            free(id);
            return -1;
        };

        printf(_("Message `%s' accepted\n"), argv[i]);
        free(id);
    }

    if (argc < 2) {
        /* TODO: list commercial not received messages and accept all of them
         * */
        printf(_("Error: No message ID supplied. Accepting all commercial "
                    "messages not implemented yet.\n"));
        return -1;
    }

    return 0;
}


/* Mark message as read. At one form of ID must be provided.
 * @id is UTF-8 encoded message ID
 * @id_locale is locale encoded message ID. @id takes preference. */
static int do_read_message(const char *id, const char *id_locale) {
    _Bool static_id;
    isds_error err;

    if ((!id || !*id) && (!id_locale || !*id_locale)) return -1;

    if (id) {
        static_id = 1;
        id_locale = utf82locale(id);
    } else {
        static_id = 0;
        id = locale2utf8(id_locale);
        if (!id) {
            printf(_("Error: Could not convert message ID to UTF-8: %s\n"),
                    id_locale);
            return -1;
        }
    }

    printf(_("Marking message `%s' as read...\n"), id_locale);
    err = isds_mark_message_read(cisds, id);
    finish_isds_operation(cisds, err);

    if (!err)
        printf(_("Message `%s' marked as read\n"), id_locale);

    if (static_id) free((char *) id_locale);
    else free((char *) id);

    return (err) ? -1 : 0;
}


static void shi_read_message_usage(const char *command) {
    printf(_(
"Usage: %s [MESSAGE_ID...]\n"
"Mark message as read moving its state to read.\n"
"\n"
"When new incoming message is download, its state is not changed on server.\n"
"Client must mark such message as read explicitly. You can use this command\n"
"to do so, if not done automatically at download time by your client.\n"
"\n"
"If MESSAGE_ID is defined, mark that message. More messages can be specified.\n"
"Otherwise marks currently loaded message.\n"),
            command);
}


static int shi_read_message(int argc, const char **argv) {
    for (int i = 1; i < argc; i++) {
        if (!argv[i] || !*argv[i]) continue;
        if (do_read_message(NULL, argv[i]))
            return -1;
    }

    if (argc < 2) {
        if (!message) { 
            printf(_("No message loaded\n"));
            return -1;
        }
        if (!message->envelope) { 
            printf(_("Loaded message is missing envelope\n"));
            return -1;
        }
        if (!message->envelope->dmID || !*message->envelope->dmID) { 
            printf(_("Loaded message is missing ID\n"));
            return -1;
        }
        return do_read_message(message->envelope->dmID, NULL);
    }

    return 0;
}


static int shi_show_message(int argc, const char **argv) {
    if (!message) {
        printf(_("No message loaded\n"));
        return -1;
    }

    format_message(message);
    return 0;
}


static void shi_incoming_message_usage(const char *command) {
    printf(_(
                "Usage: %s MESSAGE_ID\n"
                "Get incoming message with MESSAGE_ID.\n"),
            command);
}


static int shi_incoming_message(int argc, const char **argv) {
    isds_error err;
    const char *id;

    if (!argv || !argv[1] || !*argv[1]) {
        shi_incoming_message_usage(argv[0]);
        return -1;
    }
    id = argv[1];

    printf(_("Getting incoming message...\n"));
    err = isds_get_signed_received_message(cisds, id, &message);
    finish_isds_operation(cisds, err);
    if (err) {
        set_prompt(NULL);
        select_completition(COMPL_COMMAND);
        return -1;
    }

    format_message(message);

    if (message->envelope && message->envelope->dmID)
        set_prompt(_("%s %s"), argv[0], message->envelope->dmID);
    else 
        set_prompt("%s", argv[0]);
    select_completition(COMPL_MSG);
    return 0;
}


static void shi_outgoing_message_usage(const char *command) {
    printf(_(
                "Usage: %s MESSAGE_ID\n"
                "Get outgoing message with MESSAGE_ID.\n"),
            command);
}


static int shi_outgoing_message(int argc, const char **argv) {
    isds_error err;
    const char *id;

    if (!argv || !argv[1] || !*argv[1]) {
        shi_outgoing_message_usage(argv[0]);
        return -1;
    }
    id = argv[1];

    printf(_("Getting outgoing message...\n"));
    err = isds_get_signed_sent_message(cisds, id, &message);
    finish_isds_operation(cisds, err);
    if (err) {
        set_prompt(NULL);
        select_completition(COMPL_COMMAND);
        return -1;
    }

    format_message(message);
    if (message->envelope && message->envelope->dmID)
        set_prompt(_("%s %s"), argv[0], message->envelope->dmID);
    else 
        set_prompt("%s", argv[0]);
    select_completition(COMPL_MSG);
    return 0;
}


static void shi_load_anything_usage(const char *command) {
    printf(_(
                "Usage: %s FILE\n"
                "Load message or message delivery details from local FILE.\n"),
            command);
}


static int shi_load_anything(int argc, const char **argv) {
    int fd;
    void *buffer = NULL;
    size_t length;
    isds_raw_type raw_type;
    isds_error err;
    char *type_name = NULL;

    if (!argv || !argv[1] || !*argv[1]) {
        shi_load_anything_usage((argv)?argv[0]:NULL);
        return -1;
    }

    printf(_("Loading file `%s'...\n"), argv[1]);

    if (mmap_file(argv[1], &fd, &buffer, &length)) return -1;

    printf(_("Detecting file format...\n"));
    err = isds_guess_raw_type(cisds, &raw_type, buffer, length);
    finish_isds_operation(cisds, err);
 
    if (err) {
        if (err == IE_NOTSUP)
            printf(_("Unknown format. Could not parse the file.\n"));
        else 
            printf(_("Error while detecting format. "
                        "Could not parse the file.\n"));
    } else {
        switch (raw_type) {
            case RAWTYPE_INCOMING_MESSAGE:
            case RAWTYPE_PLAIN_SIGNED_INCOMING_MESSAGE:
            case RAWTYPE_CMS_SIGNED_INCOMING_MESSAGE:
            case RAWTYPE_PLAIN_SIGNED_OUTGOING_MESSAGE:
            case RAWTYPE_CMS_SIGNED_OUTGOING_MESSAGE:
                err = isds_load_message(cisds, raw_type,
                        buffer, length, &message, BUFFER_COPY);
                finish_isds_operation(cisds, err);
                type_name = N_("message");
                break;

            case RAWTYPE_DELIVERYINFO:
            case RAWTYPE_PLAIN_SIGNED_DELIVERYINFO:
            case RAWTYPE_CMS_SIGNED_DELIVERYINFO:
                err = isds_load_delivery_info(cisds, raw_type,
                        buffer, length, &message, BUFFER_COPY);
                finish_isds_operation(cisds, err);
                type_name = N_("delivery");
                break;

            default:
                printf(_("Unsupported format. Could not parse the file.\n"));
                err = IE_NOTSUP;
        }
    }

    munmap_file(fd, buffer, length);

    if (err) {
        set_prompt(NULL);
        select_completition(COMPL_COMMAND);
        return -1;
    }

    format_message(message);

    if (message->envelope && message->envelope->dmID)
        set_prompt(_("%s %s"), _(type_name), message->envelope->dmID);
    else 
        set_prompt("%s", _(type_name));
    select_completition(COMPL_MSG);
    return 0;
}


static void shi_save_message_usage(const char *command) {
    printf(_(
                "Usage: %s FILE\n"
                "Save message into local FILE.\n"),
            command);
}


static const char *raw_type2mime(isds_raw_type raw_type) {
    switch (raw_type) {
        case RAWTYPE_INCOMING_MESSAGE:
        case RAWTYPE_PLAIN_SIGNED_INCOMING_MESSAGE:
        case RAWTYPE_PLAIN_SIGNED_OUTGOING_MESSAGE:
        case RAWTYPE_DELIVERYINFO:
        case RAWTYPE_PLAIN_SIGNED_DELIVERYINFO:
            return "text/xml";

        case RAWTYPE_CMS_SIGNED_INCOMING_MESSAGE:
        case RAWTYPE_CMS_SIGNED_OUTGOING_MESSAGE:
        case RAWTYPE_CMS_SIGNED_DELIVERYINFO:
            return "application/pkcs7-mime";

        default:
            return NULL;
    }
}


static int shi_save_message(int argc, const char **argv) {
    if (!argv || !argv[1] || !*argv[1]) {
        shi_save_message_usage(argv[0]);
        return -1;
    }

    if (!message) {
        printf(_("No message loaded\n"));
        return -1;
    }
    if (!message->raw || message->raw_length == 0) {
        printf(_("Loaded message is missing raw representation\n"));
        return -1;
    }

    return save_data_to_file(argv[1], message->raw, message->raw_length,
            raw_type2mime(message->raw_type));
}


/* Return document of current message identified by ordinal number expressed
 * as string. In case of error return NULL. */
static const struct isds_document *locate_document_by_ordinal_string(
        const char *number) {
    const struct isds_list *item;
    const struct isds_document *document = NULL;
    int ordinar, i;

    if (!number) return NULL;

    ordinar = atoi(number);
    if (ordinar <= 0) {
        printf(_("%s: Document number must be positive number\n"), number);
        return NULL;
    }

    if (!message) {
        printf(_("No message loaded\n"));
        return NULL;
    }

    /* Find document */
    for (item = message->documents, i = 0; item; item = item->next) {
        if (!item->data) continue;
        if (++i == ordinar) {
            document = (const struct isds_document *) item->data;
            break;
        }
    }
    if (i != ordinar) {
        printf(_("Message does not contain document #%d\n"), ordinar);
        return NULL;
    }

    return document;
}


static void shi_save_document_usage(const char *command) {
    printf(_(
"Usage: %s NUMBER [DESTINATION]\n"
"Save document having ordinal NUMBER within current message into local file.\n"
"If DESTINATION is file (or does not exist yet), document will be saved into\n"
"this file.\n"
"If DESTINATION is existing directory, file name equaled to document name\n"
"will be saved into DESTINATION.\n"
"If DESTINATION is missing, document name will be used as file name and\n"
"saved into working directory.\n"
"Be aware that document name does not embed malicious characters (slashes).\n"
"\n"
"If the document is a binary stream, image of the document will be copied\n"
"into a file. If the document is a XML document, the XML tree will be serialized\n"
"into a file. If XML document stands for one element or one text node, the node\n"
"(and its children recursively) will be serialized. If XML document compounds\n"
"more nodes or a comment or a processing instruction, parent node from ISDS name\n"
"space will be used to ensure output serialized XML well-formness.\n"
),
            command);
}


static int shi_save_document(int argc, const char **argv) {
    const struct isds_document *document;
    const char *dirname = NULL;
    char *filename = NULL, *path = NULL;
    int retval = 0;

    if (!argv || !argv[1] || !*argv[1] || argc > 3) {
        shi_save_document_usage(argv[0]);
        return -1;
    }

    document = locate_document_by_ordinal_string(argv[1]);
    if (!document) return -1;

    /* Select directory and file name */
    if (argv[2] && *argv[2]) {
        if (!is_directory(argv[2])) {
            dirname = argv[2];
        } else {
            filename = strdup(argv[2]);
            if (!filename) {
                printf(_("Not enough memory\n"));
                return -1;
            }
        }
    }
    if (!filename && document->dmFileDescr && &document->dmFileDescr) {
        filename = utf82locale(document->dmFileDescr);
        if (!filename) {
            printf(_("Not enough memory\n"));
            return -1;
        }
    }
    if (!filename) {
        printf(_("File name neither supplied, nor document name exists\n"
                    "Please, supply one.\n"));
        return -1;
    }

    /* Build path */
    if (dirname) {
        path = astrcat3(dirname, "/", filename);
        zfree(filename);
    } else {
        path = filename;
        filename = NULL;
    }
    if (!path) {
        printf(_("Not enough memory\n"));
        return -1;
    }

    /* Save document */ 
    if (document->is_xml)
        retval = save_xml_to_file(path, document->xml_node_list,
                document->dmMimeType);
    else
        retval = save_data_to_file(path, document->data, document->data_length,
                document->dmMimeType);
    free(path);
    return retval;
}


static void shi_save_stamp_usage(const char *command) {
    printf(_(
                "Usage: %s FILE\n"
                "Save message time stamp into local FILE.\n"),
            command);
}


static int shi_save_stamp(int argc, const char **argv) {
    if (!argv || !argv[1] || !*argv[1]) {
        shi_save_message_usage(argv[0]);
        return -1;
    }

    if (!message) {
        printf(_("No message loaded\n"));
        return -1;
    }
    if (!message->envelope || !message->envelope->timestamp||
            message->envelope->timestamp_length == 0) {
        printf(_("Loaded message is missing time stamp\n"));
        return -1;
    }

    return save_data_to_file(argv[1],
            message->envelope->timestamp, message->envelope->timestamp_length,
            "application/timestamp-reply");
}


static int shi_show_list(int argc, const char **argv) {
    if (!messages) {
        printf(_("No message list loaded\n"));
        return -1;
    }

    printf((messages_are_outgoing) ?
            ngettext("You have %'lu outgoing message\n",
                "You have %'lu outgoing messages\n", total_messages) :
            ngettext("You have %'lu incoming message\n",
                "You have %'lu incoming messages\n", total_messages),
            total_messages);
    print_message_list(messages, messages_are_outgoing);
    return 0;
}


static int shi_list_incoming(int argc, const char **argv) {
    isds_error err;

    printf(_("Listing incoming messages...\n"));
    err = isds_get_list_of_received_messages(cisds, NULL, NULL, NULL,
            MESSAGESTATE_ANY, 0, &total_messages, &messages);
    finish_isds_operation(cisds, err);
    if (err) {
        set_prompt(NULL);
        select_completition(COMPL_COMMAND);
        return -1;
    }
    messages_are_outgoing = 0;

    shi_show_list(0, NULL);

    set_prompt(_("%s %'lu"), argv[0], total_messages);
    select_completition(COMPL_LIST);
    return 0;
}


static int shi_list_outgoing(int argc, const char **argv) {
    isds_error err;

    printf(_("Listing outgoing messages...\n"));
    err = isds_get_list_of_sent_messages(cisds, NULL, NULL, NULL,
            MESSAGESTATE_ANY, 0, &total_messages, &messages);
    finish_isds_operation(cisds, err);
    if (err) {
        set_prompt(NULL);
        select_completition(COMPL_COMMAND);
        return -1;
    }
    messages_are_outgoing = 1;

    shi_show_list(0, NULL);

    set_prompt(_("%s %'lu"), argv[0], total_messages);
    select_completition(COMPL_LIST);
    return 0;
}


/* Submit document for conversion and print assigned identifier */
static int do_convert(const struct isds_document *document) {
    isds_error err;
    char *id = NULL;
    struct tm *date = NULL;

    if (!document) return -1;

    printf(_("Submitting document for authorized conversion...\n"));

    err = czp_convert_document(czechpoint, document, &id, &date);
    finish_isds_operation(czechpoint, err);

    if (!err) {
        char *name_locale = utf82locale(document->dmFileDescr);
        char *date_string = tm2string(date);
        char *id_locale = utf82locale(id);
        printf(_(
"Document submitted for authorized conversion successfully under name\n"
"`%s' on %s.\n"
"Submit identifier assigned by Czech POINT deposit is `%s'.\n"),
                name_locale, date_string, id_locale);
        free(name_locale);
        free(date_string);
        free(id_locale);
        printf(_("Be ware that submitted document has restricted lifetime "
                    "(30 days).\n"));
        printf(_("See <%s> for more details.\n"), CZPDEPOSIT_URL);
    }

    free(id); free(date);
    return (err) ? -1 : 0;
}


static void shi_convert_file_usage(const char *command) {
    printf(_(
"Usage: %s FILE [NAME]\n"
"Submit local FILE to authorized conversion under NAME. If NAME is missing,\n"
"it will use FILE name.\n"), command);
    printf(_(
"\n"
"If Czech POINT deposit accepts document, it will return document identifier\n"
"that user is supposed to provide to officer at Czech POINT contact place.\n"
"Currently only PDF 1.3 and higher version files are accepted.\n"));
    printf(_("See <%s> for more details.\n"), CZPDEPOSIT_URL);
}


static int shi_convert_file(int argc, const char **argv) {
    int fd;
    struct isds_document document;
    int retval = 0;

    if (!argv || argc > 3 || !argv[1] || !*argv[1]) {
        shi_convert_file_usage((argv)?argv[0]:NULL);
        return -1;
    }

    memset(&document, 0, sizeof(document));

    if (argc == 3 && argv[2] && *argv[2])
        document.dmFileDescr = locale2utf8(argv[2]);
    else
        document.dmFileDescr = locale2utf8(argv[1]);
    if (!document.dmFileDescr) {
        printf(_("Could not convert document name to UTF-8\n"));
        return -1;
    }

    printf(_("Loading document from file `%s'...\n"), argv[1]);
    if (mmap_file(argv[1], &fd, &document.data, &document.data_length)) {
        free(document.dmFileDescr);
        return -1;
    }

    retval = do_convert(&document);

    munmap_file(fd, document.data, document.data_length);
    free(document.dmFileDescr);
    return retval;
}


static void shi_convert_document_usage(const char *command) {
    printf(_(
"Usage: %s NUMBER\n"
"Submit message document with ordinal NUMBER to authorized conversion.\n"),
            command);
    printf(_(
"\n"
"If Czech POINT deposit accepts document, it will return document identifier\n"
"that user is supposed to provide to officer at Czech POINT contact place.\n"
"Currently only PDF 1.3 and higher version files are accepted.\n"));
    printf(_("See <%s> for more details.\n"), CZPDEPOSIT_URL);
}


static int shi_convert_document(int argc, const char **argv) {
    const struct isds_document *document;

    if (!argv || !argv[1] || !*argv[1]) {
        shi_convert_document_usage((argv)?argv[0]:NULL);
        return -1;
    }

    document = locate_document_by_ordinal_string(argv[1]);
    if (!document) return -1;

    return do_convert(document);
}


#if ENABLE_DEBUG
static void shi_print_usage(const char *command) {
    printf(_(
"Usage: %s STRING LENGTH\n"
"Prints STRING into LENGTH columns. Negative LENGTH means not to cut\n"
"overflowing string.\n"
"This should be locale and terminal agnostic.\n"),
            command);
}


static int shi_print(int argc, const char **argv) {
    long int width;

    if (!argv || !argv[1] || !argv[2]) {
        shi_print_usage((argv)?argv[0]:NULL);
        return -1;
    }

    width = strtol(argv[2], NULL, 10);
    if (width < INT_MIN) {
        fprintf(stderr,
            _("Length argument must not lesser than %d.\n"), INT_MIN);
        return -1;
    }
    if (width > INT_MAX) {
        fprintf(stderr,
            _("Length argument must not be greater than %d.\n"), INT_MAX);
        return -1;
    }

    printf(_(">"));
    fnprint(stdout, argv[1], width);
    printf(_("<\n"));

    return 0;
}


static int shi_tokenize(int argc, const char **argv) {

    if (!argv) return 0;

    for (int i = 0; i < argc; i++) {
        printf(_(">%s<\n"), argv[i]);
    }
    return 0;
}


static int shi_quote(int argc, const char **argv) {
    char *escaped, *unescaped;

    if (!argv) return 0;

    printf(_("Original\tQuoted\tDequoted\n"));
    for (int i = 0; i < argc; i++) {
        escaped = shi_quote_filename((char *) argv[i], 0, NULL);
        unescaped = shi_dequote_filename((char *) argv[i], 0);
        printf(_(">%s<\t>%s<\t>%s<\n"), argv[i], escaped, unescaped);
        free(escaped);
        free(unescaped);
    }
    return 0;
}
#endif


/* Interactive loop */
void shi_loop(void) {
    char *command_line = NULL;
    char **command_argv = NULL;
    int command_argc;
    struct command *command = NULL;

    printf(_("Use `help' command to get list of available commands.\n"));

    select_completition(COMPL_COMMAND);
    set_prompt(NULL);

    while (1) {
        command_line = readline((prompt) ? prompt : _("shigofumi> "));
        /* Remember not parsable commands too to user be able to get back to
         * fix command */
        if (command_line && *command_line) {
            /* TODO: Omit blank lines */
            add_history(command_line);
        }

        command_argv = tokenize(command_line, &command_argc);

        if (command_argv && command_argv[0]) {
            command = find_command(command_argv[0]);

            if (!command) {
                fprintf(stderr, _("Command not understood\n"));
            } else {
                command->function(command_argc, (const char **) command_argv);
            }
        }

        argv_free(command_argv);
        zfree(command_line);
    }
}


/* Non-interactive mode. Commands from @lines are processed until any command
 * lines remains or no error occurred. First failure terminates processing.
 * @lines is sequence of commands separated by '\n' or '\r'. The content is
 * modified during this call.
 * @return 0 if all command succeed, otherwise non-zero value
 */
int shi_batch(char *lines) {
    char *command_line;
    char **command_argv = NULL;
    int command_argc;
    struct command *command = NULL;
    int retval = 0;

    printf(_("Batch mode started.\n"));
    batch_mode = 1;
    select_completition(COMPL_COMMAND);

    while (!retval && (command_line = strtok(lines, "\n\r"))) {
        lines = NULL; /* strtok(3) requires it for subsequent calls */

        printf(_("Processing command: %s\n"), command_line);

        command_argv = tokenize(command_line, &command_argc);

        if (command_argv && command_argv[0]) {
            command = find_command(command_argv[0]);

            if (!command) {
                fprintf(stderr, _("Command not understood\n"));
                retval = -1;
            } else {
                retval = command->function(command_argc,
                        (const char **) command_argv);
            }
        }

        argv_free(command_argv);
    }

    if (retval)
        fprintf(stderr, _("Command failed!\n"));
    return retval;
}


#define COMMON_COMMANDS \
    { "accept", shi_accept_message, N_("accept commercial message"), \
        shi_accept_message_usage, ARGTYPE_MSGID }, \
    { "box", shi_box, N_("show current box details"), NULL, \
        ARGTYPE_NONE }, \
    { "cache", shi_cache, N_("show cache details"), NULL, \
        ARGTYPE_NONE }, \
    { "cd", shi_chdir, N_("change working directory"), shi_chdir_usage, \
        ARGTYPE_FILE }, \
    { "commercial", shi_commercial, \
        N_("manipulate commercial receiving box status"), \
        shi_commercial_usage, ARGTYPE_BOXID }, \
    { "compose", shi_compose, N_("compose a message"), shi_compose_usage, \
        ARGTYPE_FILE }, \
    { "convert", shi_convert_file, \
        N_("submit local document for authorized conversion"), \
        shi_convert_file_usage, ARGTYPE_FILE }, \
    { "copying", shi_copying, N_("show this program licence excerpt"), NULL, \
        ARGTYPE_NONE }, \
    { "debug", shi_debug, N_("set debugging"), shi_debug_usage, \
        ARGTYPE_FILE }, \
    { "delivery", shi_delivery, N_("get message delivery details"), \
        shi_delivery_usage, ARGTYPE_MSGID }, \
    { "findbox", shi_find_box, N_("search for a box"), shi_find_box_usage, \
        ARGTYPE_BOXID }, \
    { "hash", shi_hash, N_("query ISDS for message hash"), \
        shi_hash_usage, ARGTYPE_MSGID }, \
    { "help", shi_help, N_("describe commands"), shi_help_usage, \
        ARGTYPE_COMMAND }, \
    { "load", shi_load_anything, \
        N_("load message or message delivery details from local file"), \
        shi_load_anything_usage, ARGTYPE_FILE }, \
    { "login", shi_login, N_("log into ISDS"), NULL, ARGTYPE_NONE }, \
    { "lsi", shi_list_incoming, N_("list received messages"), NULL, \
        ARGTYPE_NONE }, \
    { "lso", shi_list_outgoing, N_("list sent messages"), NULL, \
        ARGTYPE_NONE }, \
    { "msgi", shi_incoming_message, N_("get incoming message"), \
        shi_incoming_message_usage, ARGTYPE_MSGID }, \
    { "msgo", shi_outgoing_message, N_("get outgoing message"), \
        shi_outgoing_message_usage, ARGTYPE_MSGID }, \
    { "passwd", shi_passwd, N_("manipulate user password"), shi_passwd_usage, \
        ARGTYPE_NONE }, \
    { "pwd", shi_pwd, N_("print working directory"), NULL, ARGTYPE_NONE }, \
    { "quit", shi_quit, N_("exit shigofumi"), NULL, ARGTYPE_NONE }, \
    { "read", shi_read_message, N_("mark message as read"), \
        shi_read_message_usage, ARGTYPE_MSGID }, \
    { "statbox", shi_stat_box, N_("get status of a box"), shi_stat_box_usage, \
        ARGTYPE_BOXID }, \
    { "user", shi_user, N_("show current user details"), NULL, \
        ARGTYPE_NONE }, \
    { "users", shi_users, N_("show box users"), shi_users_usage, \
        ARGTYPE_NONE }, \
    { "version", shi_version, N_("show version of this program"), NULL, \
        ARGTYPE_NONE}, \
    { NULL, NULL, NULL, NULL, ARGTYPE_NONE }

struct command base_commands[] = {
#if ENABLE_DEBUG
    { "quote", shi_quote, N_("demonstrate argument escaping"), NULL,
        ARGTYPE_FILE },
    { "print", shi_print, N_("print string into given width"),
        shi_print_usage,  ARGTYPE_NONE },
    { "tokenize", shi_tokenize, N_("demonstrate arguments tokenization"), NULL,
        ARGTYPE_FILE },
#endif
    COMMON_COMMANDS
};

struct command message_commands[] = {
    { "authenticate", shi_authenticate, N_("check message authenticity"),
        NULL, ARGTYPE_NONE },
    { "convertdoc", shi_convert_document,
        N_("submit document of current message for authorized conversion"),
        shi_convert_document_usage, ARGTYPE_DOCID },
    { "dump", shi_dump_message, N_("dump current message structure"),
        NULL, ARGTYPE_NONE },
    { "savestamp", shi_save_stamp,
        N_("save time stamp of current message into local file"),
        shi_save_stamp_usage, ARGTYPE_FILE },
    { "savedoc", shi_save_document,
        N_("save document of current message into local file"),
        shi_save_document_usage, ARGTYPE_FILE },
    { "save", shi_save_message, N_("save current message into local file"),
        shi_save_message_usage, ARGTYPE_FILE },
    { "show", shi_show_message, N_("show current message"), NULL, 
        ARGTYPE_NONE },
    { "verify", shi_verify, N_("verify current message hash"), NULL,
        ARGTYPE_NONE },
    COMMON_COMMANDS
};

struct command list_commands[] = {
    { "show", shi_show_list, N_("show current message list"), NULL, 
        ARGTYPE_NONE },
    COMMON_COMMANDS
};

#undef COMMON_COMMANDS


static void main_version(void) {
    isds_init();
    show_version();
    isds_cleanup();
    printf("\n");
    shi_copying(0, NULL);
}


static void main_usage(const char *command) {
    printf(_(
"Usage: %s [OPTION...]\n"
"Access ISDS, Process local data box messages or delivery details, submit\n"
"document to authorized conversion.\n"
"\n"
"Options:\n"
"  -c FILE      use the FILE as configuration file instead of ~/%s\n"
"  -e COMMANDS  execute COMMANDS (new line separated) and exit\n"
"  -V           show version info and exit\n"
        ),
            (command) ? command : "shigofumi",
            CONFIG_FILE);
}


int main(int argc, char **argv) {
    int opt;
    char *config_file = NULL;
    char *batch_commands = NULL;
    int retval = EXIT_SUCCESS;
    
    setlocale(LC_ALL, "");
#if ENABLE_NLS
    /* Initialize gettext */
    bindtextdomain(PACKAGE, LOCALEDIR);
    textdomain(PACKAGE);
#endif

    /* Parse arguments */
    optind = 0;
    while ((opt = getopt(argc, (char * const *)argv, "c:e:V")) != -1) {
        switch (opt) {
            case 'c':
                config_file = optarg;
                break;
            case 'e':
                batch_commands = optarg;
                break;
            case 'V':
                main_version();
                shi_exit(EXIT_SUCCESS);
                break;
            default:
                main_usage((argv[0]) ? basename(argv[0]): NULL);
                shi_exit(EXIT_FAILURE);
        }
    }


    if (shi_init(config_file)) {
        shi_exit(EXIT_FAILURE);
    }

    /*shi_login(NULL);*/

    if (batch_commands) {
        if (shi_batch(batch_commands))
            retval = EXIT_FAILURE;
    } else {
        shi_loop();
    }

    shi_exit(retval);
}
